library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

entity convolution2 is
	generic(	image_width : positive := 128;
				image_height : positive := 128;
				template_width : positive := 32;
				template_height : positive := 32);
	port (
		   image : in integer range 0 to 255;  
		   template : in integer range 0 to 255; 
		   start : in std_logic;
           image_clk : in STD_LOGIC; 
		   clk : in STD_LOGIC;
           reset : in STD_LOGIC;
		   
		   idle_indicator : out std_logic;
		   result : out integer range 0 to 2**31-1;
		   pos_x : out integer range 0 to image_width;
		   pos_y : out integer range 0 to image_height;
		   image_sum : out integer range 0 to 2**24-1;
		   image_sum_squared : out integer range 0 to 2**31-1;   
		   template_sum : out integer range 0 to 2**24-1;
		   template_sum_squared : out integer range 0 to 2**31-1;
		   result_ready : out std_logic);
end convolution2;							 

architecture Behavioral of convolution2 is	  

component RAM_module 
	generic (
		image_width : positive := image_width;
		image_height : positive := image_height;
		template_height : positive := template_height;  
		template_width : positive := template_width); 
	port (
		clk : in std_logic;
		we : in std_logic;
		addrA : in integer range 0 to image_width + template_width - 1;
		addrB : in integer range 0 to image_width + template_width - 1;
		di : in integer range 0 to 255;
		doA : out integer range 0 to 255;
		doB : out integer range 0 to 255);
end component;

subtype small_integer is integer range 0 to 255;

signal imload_u : natural range 0 to image_width - 1;	 
signal imload_v : natural range 0 to template_height - 1;
signal image_v : natural range 0 to image_height;
signal tempload_u : natural range 0 to template_width - 1;	 
signal tempload_v : natural range 0 to template_height - 1;	 
signal mem_offset : integer range 0 to template_height - 1;  --template_height
signal cur_col : integer range 0 to template_width - 1;
signal u : integer range 0 to image_width - template_width + 1;	 
signal row_wait_counter : integer range 0 to template_width;

signal result_i : integer range 0 to 2**31-1; 
signal template_sum_i : integer range 0 to 2**24-1;	   
signal template_sum_squared_i : integer range 0 to 2**31-1;  
signal image_sum_i : integer range 0 to 2**24-1;
signal image_sum_squared_i : integer range 0 to 2**31-1;   

subtype mem_array_addr_type is integer range 0 to image_width + template_width - 1;	  
type mem_array_addr is array (0 to template_height - 1) of mem_array_addr_type; 
type mem_array_data is array (0 to template_height - 1) of small_integer; 
type mem_array_we is array (0 to template_height - 1) of std_logic;

signal mem_addrA : mem_array_addr;	
signal mem_addrB : mem_array_addr;
signal mem_di : mem_array_data;
signal mem_doA : mem_array_data;  
signal mem_doB : mem_array_data;
signal mem_we : mem_array_we; 	   

signal mem_doA_shuffled : mem_array_data;  
signal mem_doB_pipeline : mem_array_data;  

type calc_array_type is array (0 to template_height - 1) of natural;
signal corr : calc_array_type;
signal i_sum : calc_array_type;
signal i_sum_squared : calc_array_type;

signal image_clk_temp : std_logic;

type state is (NOTLOADED, IDLE, IMAGE_LOAD, TEMPLATE_LOAD, PIXEL_LOAD_WAIT, RESULT_READY_1, RESULT_READY_2,
	PIXEL_LOAD, PRELOAD_DATA, CALCULATING, POST_CALCULATION_1,POST_CALCULATION_2, POST_CALCULATION_3, ROW_LOAD, ROW_WAIT, TEMPLATE_LOAD_WAIT, IMAGE_LOAD_WAIT);
signal cur_state, next_state : state;			

function calculate_row_number(row : integer) return integer is	
begin
	if (row < template_height) then				
		return row;
	else
		return row - template_height;
	end if;	
end function calculate_row_number;	 

begin 
	
	result <= result_i;   
	image_sum <= image_sum_i;
	image_sum_squared <= image_sum_squared_i;
	template_sum <= template_sum_i;
	template_sum_squared <= template_sum_squared_i;
	
	mem_array : for mem_index in 0 to template_height - 1 generate
	begin
		mem_instance : component RAM_module	 
			generic map(
				image_width => image_width,
				image_height => image_height,
				template_width => template_width,
				template_height => template_height  
				
			)
			port map (
					clk => clk,
					we => mem_we(mem_index),
					addrA => mem_addrA(mem_index),  
					addrB => mem_addrB(mem_index),
					di => mem_di(mem_index),
					doA => mem_doA(mem_index),
					doB => mem_doB(mem_index)
				);
	end generate;	
	
	edge_trigger_process : process (clk)
	begin	  
		if (rising_edge(clk)) then
			image_clk_temp <= image_clk;
		end if;
	end process edge_trigger_process;
	
	convolve : process (clk)	 
		variable corr_sum : integer range 0 to 2**31-1;  
		variable i_sum_sum : integer range 0 to 2**24-1;
		variable i_sum_squared_sum : integer range 0 to 2**31-1; 	
		type calc_array_type is array (0 to template_height - 1) of natural;
		
	begin 
		if rising_edge(clk) then
			
			if (cur_state = CALCULATING or cur_state = POST_CALCULATION_1 or cur_state = POST_CALCULATION_2 or cur_state = POST_CALCULATION_3) then

					corr_sum := 0;	
					i_sum_sum := 0;	
					i_sum_squared_sum := 0;	
					
					for i in 0 to template_height - 1 loop
						mem_doA_shuffled(i) <= mem_doA(calculate_row_number(i + mem_offset)); 
						mem_doB_pipeline(i) <= mem_doB(i);
					end loop; 	  
					
					for i in 0 to template_height - 1 loop
						corr(i) <= mem_doA_shuffled(i) * mem_doB_pipeline(i);	
						i_sum(i) <= mem_doA_shuffled(i);	 
						i_sum_squared(i) <= mem_doA_shuffled(i)*mem_doA_shuffled(i);
					end loop; 	   	   
					
					for i in 0 to template_height - 1 loop
						corr_sum := corr_sum + corr(i);	
						i_sum_sum := i_sum_sum + i_sum(i);	
						i_sum_squared_sum := i_sum_squared_sum + i_sum_squared(i);	
					end loop;
					
					result_i <= result_i + corr_sum;  
					image_sum_i <= image_sum_i + i_sum_sum;
					image_sum_squared_i <= image_sum_squared_i + i_sum_squared_sum;
					
				else  
					result_i <= 0;	
					image_sum_i <= 0;
					image_sum_squared_i <= 0;	   
					
					corr <= (others => 0);	
					i_sum <= (others => 0);	
					i_sum_squared <= (others => 0);	 
					mem_doA_shuffled <= (others => 0);	 
					
			end if;			
		end if;
	end process convolve;	  
	
	cur_col_process : process (clk)
	begin
		if rising_edge(clk) then	
			case cur_state is  
				when PRELOAD_DATA =>
					cur_col <= cur_col + 1;	
				when CALCULATING => 
					if (cur_col = template_width - 1) then
						cur_col <= 0; 
					else
						cur_col <= cur_col + 1;	
					end if;	
				when others =>	
					cur_col <= 0;
			end case;
		end if;
	end process cur_col_process;
	
	position_output_process : process (cur_state, image_v, u)
	begin
		if (cur_state = RESULT_READY_1) then
			pos_x <= u;
			pos_y <= image_v; 
		else
			pos_x <= 0;
			pos_y <= 0;
		end if;	
	end process position_output_process;
	
	
	result_ready_process : process (cur_state)
	begin
		if (cur_state = RESULT_READY_1) then
			result_ready <= '1';
		else
			result_ready <= '0';
		end if;
	end process result_ready_process;
	
	mem_offset_process : process (clk)
	begin
		if (rising_edge(clk)) then 	
			if (cur_state = PIXEL_LOAD) then 
				if (imload_u = image_width - 1) then
					if (mem_offset = template_height - 1) then 	
						mem_offset <= 0;
					else 
						mem_offset <= mem_offset + 1;
					end if;
				else
					null;
				end if;
			elsif (cur_state = IDLE or cur_state = NOTLOADED) then 
				mem_offset <= 0;
			else 
				null;
			end if;
		end if;
	end process mem_offset_process;
	
	mem_addr_process : process (cur_state, cur_col, u, tempload_u, imload_u)
	begin	
		case cur_state is
			when CALCULATING =>	  
				mem_addrA <= (others => cur_col + u);	
				mem_addrB <= (others => cur_col + image_width);	
			
			when PRELOAD_DATA =>
				mem_addrA <= (others => cur_col + u);	
				mem_addrB <= (others => cur_col + image_width);
				
			when PIXEL_LOAD =>	  
				mem_addrA <= (others => imload_u);
				mem_addrB <= (others => 0);
				
			when IMAGE_LOAD => 
				mem_addrA <= (others => imload_u);
				mem_addrB <= (others => 0);
			
			when ROW_WAIT =>	
				mem_addrA <= (others => imload_u);
			   	mem_addrB <= (others => 0);
				
			when TEMPLATE_LOAD => 
				mem_addrA <= (others => image_width + tempload_u);
				mem_addrB <= (others => 0);

			when others =>
				mem_addrA <= (others => 0);
				mem_addrB <= (others => 0);
		end case; 
	end process mem_addr_process;
	
	ucounter_process : process (clk)
	begin
		if (rising_edge(clk)) then
			if (cur_state = PIXEL_LOAD) then
				if (u < image_width - template_width) then
					u <= u + 1;
				else 
					u <= 0;
				end if;
			elsif (cur_state = IDLE) then
				u <= 0;
			else 
				null;
			end if;
		end if;
	end process ucounter_process;

	imloaduv_counters : process (clk)
	begin			   
		if (rising_edge(clk)) then	
			if (cur_state = PIXEL_LOAD or cur_state = IMAGE_LOAD or cur_state = ROW_LOAD) then
				if (imload_u = image_width - 1) then
					imload_u <= 0;
				else 
					imload_u <= imload_u + 1;
				end if;
				
				if (image_v = image_height and imload_u = image_width - 1) then
					image_v <= 0;
				elsif (imload_u = image_width - 1) then
						image_v <= image_v + 1;			
				else
					null;
				end if;
				
				if (imload_v = template_height - 1 and imload_u = image_width - 1) then
					imload_v <= 0;
				elsif (imload_u = image_width - 1 ) then
					imload_v <= imload_v + 1;
				else
					null;
				end if;	
			elsif (cur_state = IDLE) then
				imload_u <= 0;
				imload_v <= 0;	
				image_v <= 0;
			else
				null;
			end if;	 
		end if;
	end process imloaduv_counters;
	
	temploaduv_counters : process (clk)
	begin			   
		if (rising_edge(clk)) then	
			if (cur_state = TEMPLATE_LOAD) then
				if (tempload_u = template_width - 1) then
					tempload_u <= 0;
				else 
					tempload_u <= tempload_u + 1;
				end if;
				
				if (tempload_v = template_height - 1 and tempload_u = template_width - 1) then
					tempload_v <= 0;
				elsif (tempload_u = template_width - 1 ) then
					tempload_v <= tempload_v + 1;
				else
					null;
				end if;	
			elsif (cur_state = IDLE) then
				tempload_u <= 0;
				tempload_v <= 0;
			else
				null;
			end if;	 
		end if;
	end process temploaduv_counters;
	
	row_wait_counter_proc : process (image_clk)
	begin	  
		if (rising_edge(image_clk)) then
			case cur_state is 
				when ROW_WAIT =>
					row_wait_counter <= row_wait_counter + 1;
				when others =>
					row_wait_counter <= 0;
			end case;		
		end if;
	end process row_wait_counter_proc;
	
	load_image : process (cur_state, imload_v,image,template,tempload_v)
	begin
		if (cur_state = IMAGE_LOAD or cur_state = PIXEL_LOAD or cur_state = ROW_WAIT) then	  
			mem_di <= (others => 0);	 
			mem_di(calculate_row_number(imload_v)) <= image; 		
			
			mem_we <= (others => '0');	
			mem_we(calculate_row_number(imload_v)) <= '1';	 
		elsif (cur_state = TEMPLATE_LOAD) then 
			mem_di <= (others => 0);
			mem_di(tempload_v) <= template; 
			
			mem_we <= (others => '0');	
			mem_we(tempload_v) <= '1';	
		else   
			mem_di <= (others => 0);
			mem_we <= (others => '0');
		end if;
	end process load_image;	 
	
	template_sum_process : process (clk)
	begin
		if (rising_edge(clk)) then 								 
			if (cur_state = TEMPLATE_LOAD) then
				template_sum_i <= template_sum_i + template;
				template_sum_squared_i <= template_sum_squared_i + template * template;			
			elsif (cur_state = NOTLOADED) then
				template_sum_i <= 0;
				template_sum_squared_i <= 0;			
			else 
				null;
			end if;
		end if;
	end process template_sum_process;	 
	
	convolution_FSM_sync : process (clk)
	begin  
		if (rising_edge(clk)) then
			if (reset = '1') then
				cur_state <= NOTLOADED;
			else
				cur_state <= next_state;
			end if;
		end if;
	end process convolution_FSM_sync;
	
	idle_proc : process (cur_state)
	begin
		if (cur_state = IDLE) then
			idle_indicator <= '1';
		else
			idle_indicator <= '0';
		end if;
	end process idle_proc;

	convolution_FSM_comb : process (image_v, cur_state, cur_col, start, image_clk, image_clk_temp, 
		row_wait_counter, imload_u, imload_v, tempload_u, tempload_v, u)
	begin  
		case (cur_state) is
			when NOTLOADED => 
			   	if (start = '1') then
					next_state <= TEMPLATE_LOAD_WAIT;
				else
					next_state <= NOTLOADED;
				end if;	
				
			when IDLE =>
				if (start = '1') then
					next_state <= IMAGE_LOAD_WAIT;
				else
					next_state <= IDLE;
				end if;
				
			when TEMPLATE_LOAD_WAIT =>
				if (image_clk = '1' and image_clk_temp = '0') then
					next_state <= TEMPLATE_LOAD;
				else 
					next_state <= TEMPLATE_LOAD_WAIT;
				end if;

			when TEMPLATE_LOAD =>
				if (tempload_v = template_height - 1 and tempload_u = template_width - 1) then --TODO
					next_state <= IMAGE_LOAD_WAIT;
				else 
					next_state <= TEMPLATE_LOAD_WAIT;
				end if;	
			
			when IMAGE_LOAD_WAIT =>
				if (image_clk = '1' and image_clk_temp = '0') then
					next_state <= IMAGE_LOAD;
				else 
					next_state <= IMAGE_LOAD_WAIT;
				end if;

			when IMAGE_LOAD =>
				if (imload_u = image_width - 1 and imload_v = template_height - 1 ) then
					next_state <= PRELOAD_DATA;
				else 
					next_state <= IMAGE_LOAD_WAIT;
				end if;	
						
			when PIXEL_LOAD_WAIT =>
				if (image_clk = '1' and image_clk_temp = '0') then
					next_state <= PIXEL_LOAD;
				else 
					next_state <= PIXEL_LOAD_WAIT;
				end if;

			when PIXEL_LOAD =>
				if (imload_u = image_width - 1 and image_v = image_height) then
					next_state <= IDLE;
				else 
					next_state <= PRELOAD_DATA;
				end if;	
		
			when PRELOAD_DATA =>
				next_state <= CALCULATING;
				
			when CALCULATING =>
				if (cur_col = template_width - 1) then
					next_state <= POST_CALCULATION_1;
				else 
					next_state <= CALCULATING;
				end if;				
				
			when POST_CALCULATION_1 =>  
				next_state <= POST_CALCULATION_2;
			
			when POST_CALCULATION_2 =>  
				next_state <= POST_CALCULATION_3;	
				
			when POST_CALCULATION_3 =>  
				next_state <= RESULT_READY_1;	
				
			when ROW_WAIT =>
				if (image_clk = '1' and image_clk_temp = '0') then
					next_state <= ROW_LOAD;
				else 
					next_state <= ROW_WAIT;
				end if;	 
				
			when ROW_LOAD =>
				if (row_wait_counter = template_width - 1) then
					next_state <= PIXEL_LOAD_WAIT;
				else 
					next_state <= ROW_WAIT;
				end if;
				
			when RESULT_READY_1 =>
				next_state <= RESULT_READY_2;	 
				
			when RESULT_READY_2 =>
				if (u = image_width - template_width ) then
						next_state <= ROW_WAIT;
					else	
						next_state <= PIXEL_LOAD_WAIT;	
				end if;	  

		end case;		
	end process convolution_FSM_comb;	
		
end Behavioral;									   
